@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 # BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
-# Copyright (C) 2020-2021, Yoel Cortes-Pena <yoelcortes@gmail.com>
+# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
 # 
 # This module is under the UIUC open-source license. See 
 # github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
@@ -11,7 +11,7 @@ import numpy as np
 import flexsolve as flx
 from .fugacity_coefficients import IdealFugacityCoefficients
 from .domain import vle_domain
-from ..exceptions import InfeasibleRegion, DomainError
+from ..exceptions import InfeasibleRegion
 from .. import functional as fn
 from ..utils import fill_like, Cache
 from .._settings import settings
@@ -114,21 +114,22 @@ class BubblePoint:
     
     def _T_error(self, T, P, z_over_P, z_norm, y):
         if T <= 0: raise InfeasibleRegion('negative temperature')
+        Psats = np.array([i(T) for i in self.Psats], dtype=float)
         y_phi =  (z_over_P
-                  * np.array([i(T) for i in self.Psats])
+                  * Psats
                   * self.gamma(z_norm, T) 
-                  * self.pcf(z_norm, T))
+                  * self.pcf(T, P, Psats))
         y[:] = solve_y(y_phi, self.phi, T, P, y)
         return 1. - y.sum()
     
-    def _P_error(self, P, T, z_Psat_gamma_pcf, y):
+    def _P_error(self, P, T, z_Psat_gamma, Psats, y):
         if P <= 0: raise InfeasibleRegion('negative pressure')
-        y_phi = z_Psat_gamma_pcf / P
+        y_phi = z_Psat_gamma * self.pcf(T, P, Psats) / P
         y[:] = solve_y(y_phi, self.phi, T, P, y)
         return 1. - y.sum()
         
     def _T_error_ideal(self, T, z_over_P, y):
-        y[:] = z_over_P * np.array([i(T) for i in self.Psats])
+        y[:] = z_over_P * np.array([i(T) for i in self.Psats], dtype=float)
         return 1 - y.sum()
     
     def _Ty_ideal(self, z_over_P):
@@ -197,7 +198,8 @@ class BubblePoint:
         if N == 0:
             raise ValueError('no components present')
         if N == 1:
-            T = self.chemicals[fn.first_true_index(positives)].Tsat(P, check_validity=False)
+            chemical = self.chemicals[fn.first_true_index(positives)]
+            T = chemical.Tsat(P, check_validity=False) if P <= chemical.Pc else chemical.Tc
             y = z.copy()
         else:
             f = self._T_error
@@ -209,7 +211,7 @@ class BubblePoint:
                 T = flx.aitken_secant(f, T_guess, T_guess + 1e-3,
                                       self.T_tol, 5e-12, args,
                                       checkiter=False)
-            except (InfeasibleRegion, DomainError):
+            except RuntimeError:
                 Tmin = self.Tmin; Tmax = self.Tmax
                 T = flx.IQ_interpolation(f, Tmin, Tmax,
                                          f(Tmin, *args), f(Tmax, *args),
@@ -251,21 +253,22 @@ class BubblePoint:
         if N == 0:
             raise ValueError('no components present')
         if N == 1:
-            P = self.chemicals[fn.first_true_index(positives)].Psat(T)
+            chemical = self.chemicals[fn.first_true_index(positives)]
+            P = chemical.Psat(T) if T <= chemical.Tc else chemical.Pc
             y = z.copy()
         else:
             if T > self.Tmax: T = self.Tmax
             elif T < self.Tmin: T = self.Tmin
             Psat = np.array([i(T) for i in self.Psats])
             z_norm = z / z.sum()
-            z_Psat_gamma_pcf = z * Psat * self.gamma(z_norm, T) * self.pcf(z_norm, T)
+            z_Psat_gamma = z * Psat * self.gamma(z_norm, T)
             f = self._P_error
-            P_guess, y = self._Py_ideal(z_Psat_gamma_pcf)
-            args = (T, z_Psat_gamma_pcf, y)
+            P_guess, y = self._Py_ideal(z_Psat_gamma)
+            args = (T, z_Psat_gamma, Psat, y)
             try:
                 P = flx.aitken_secant(f, P_guess, P_guess-1, self.P_tol, 1e-9,
                                       args, checkiter=False)
-            except (InfeasibleRegion, DomainError):
+            except RuntimeError:
                 Pmin = self.Pmin; Pmax = self.Pmax
                 P = flx.IQ_interpolation(f, Pmin, Pmax,
                                          f(Pmin, *args), f(Pmax, *args),
