@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 # BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
-# Copyright (C) 2020-2021, Yoel Cortes-Pena <yoelcortes@gmail.com>
+# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
 # 
 # This module is under the UIUC open-source license. See 
 # github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
@@ -9,17 +9,31 @@
 """
 from .utils import colors
 
-__all__ = ('UndefinedChemicalAlias',
-           'UndefinedChemical',
-           'UndefinedPhase',
-           'DimensionError',
-           'InfeasibleRegion',
-           'DomainError',
-           'InvalidMethod',
-           'NoEquilibrium',
-           'message_with_object_stamp',
-           'try_method_with_object_stamp',
-           'raise_error_with_object_stamp')
+__all__ = (
+    'StampedKeyError',
+    'UndefinedChemicalAlias',
+    'UndefinedChemical',
+    'UndefinedPhase',
+    'DimensionError',
+    'InfeasibleRegion',
+    'InvalidMethod',
+    'NoEquilibrium',
+    'message_with_object_stamp',
+    'try_method_with_object_stamp',
+    'raise_error_with_object_stamp'
+)
+
+class StampedKeyError(KeyError):
+    def __init__(self, msg):
+        super().__init__(msg)
+    
+    def __repr__(self):
+        return f"KeyError({self.args[0]})"
+    
+    def __str__(self):
+        return self.args[0]
+
+StampedKeyError.__name__ = 'KeyError'
 
 class InfeasibleRegion(RuntimeError):
     """Runtime error regarding infeasible processes."""
@@ -46,12 +60,6 @@ class NoEquilibrium(Exception):
 class DimensionError(ValueError):
     """ValueError regarding wrong dimensions."""
 
-class DomainError(ValueError):
-    """ValueError regarding an attempt to evaluate a model out of its domain."""
-    def __init__(self, msg, **data):
-        super().__init__(msg)
-        self.__dict__.update(data)
-
 class InvalidMethod(ValueError):
     """ValueError regarding an attempt to evaluate an invalid method."""
     def __init__(self, method):
@@ -74,7 +82,9 @@ def raise_error_with_object_stamp(object, error):
 def try_method_with_object_stamp(object, method, args=()):
     try:
         return method(*args)
+    except StampedKeyError as error:
+        raise StampedKeyError(message_with_object_stamp(object, error.args[0]))
     except KeyError as error:
-        raise error
+        raise StampedKeyError(message_with_object_stamp(object, repr(error.args[0])))
     except Exception as error:
         raise_error_with_object_stamp(object, error)
\ No newline at end of file
