@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 # BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
-# Copyright (C) 2020-2021, Yoel Cortes-Pena <yoelcortes@gmail.com>
+# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
 # 
 # This module is under the UIUC open-source license. See 
 # github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
@@ -13,6 +13,7 @@ __all__ = ('registered', 'unregistered', 'registered_franchise')
 
 def unregistered(cls):
     cls._register = _pretend_to_register
+    cls.register_alias = _pretend_to_register
     return cls
 
 def registered(ticket_name):
@@ -27,6 +28,7 @@ def _registered(cls, ticket_name, autonumber=True):
     cls._take_unregistered_ticket = _take_unregistered_ticket
     cls._take_ticket = _take_ticket
     cls._register = _register
+    cls.register_alias = register_alias
     cls.ID = ID
     cls.__repr__ = __repr__
     cls.__str__ = __str__
@@ -44,7 +46,7 @@ def _take_ticket(cls):
 
 @classmethod
 def _take_unregistered_ticket(cls):
-    if cls.unregistered_ticket_number > 100:
+    if cls.unregistered_ticket_number > 1000:
        cls.unregistered_ticket_number = 1 
     else:
         cls.unregistered_ticket_number += 1
@@ -82,8 +84,14 @@ def ID(self):
 def ID(self, ID):
     self._register(ID)
     
+def register_alias(self, alias):
+    self.registry.register_alias_safely(alias, self) 
+    
 def __repr__(self):
-    return f"<{type(self).__name__}: {self.ID}>"
+    if self.ID:
+        return f'<{type(self).__name__}: {self.ID}>'
+    else:
+        return f'<{type(self).__name__}>'
 
 def __str__(self):
     return self.ID or "."
